import numpy as np
import sympy as sp
import warnings
from ..utils.Initialize import Space, Problem_type, Get_constraints, x_Generator, Get_x_and_z, Bounds, Get_kernel, Points_mesh, Percentile_q, Num_jobs, AF_params
from ..utils.Dimension_reduction import Reduce, Inverse, Train_reducer, Train_inverter, Find_reducer, Find_inverter, Red_bounds
from ..utils.Models import Select_model, Train_model, Train_models_const
from ..utils.Acq_fun import AF
from ..utils.Set_level_filtration import Slf
from ..utils.Querry_points import Querry
from ..utils.Update import Up_mesh
from ..utils.Aux import Errors, Data_eval, Eval_fun, Eval_const, Best_values, Regret, Print_header, Print_results, Create_results

class BO():

    def __init__(self, function, domain, sense,
                 surrogate = "GP",
                 acquisition_function = "UCB",
                 xi = 2,
                 xi_decay = "yes",
                 kernel = None,
                 kern_discovery = "yes",
                 kern_discovery_evals = 2,
                 x_0 = None, 
                 z_0 = None,
                 design = "LHS",
                 p_design = None,
                 n_jobs = None,
                 max_iter = 100, 
                 n_restarts = 5,
                 constraints = None,
                 constraints_method = "PoF",
                 reducer = None,
                 n_components = None,
                 inverter_transform = "No",
                 verbose = 0,
                 ):

        self.function = function
        self.domain = domain
        self.constraints = constraints
        self.x_0 = x_0
        self.z_0 = z_0
        self.sense = sense
        self.surrogate = surrogate
        self.acquisition_function = acquisition_function
        self.kernel = kernel
        self.n_jobs = n_jobs
        self.kern_discovery = kern_discovery
        self.design = design
        self.p_design = p_design
        self.max_iter = max_iter
        self.n_restarts = n_restarts
        self.kern_discovery_evals = kern_discovery_evals
        self.xi = xi
        self.xi_decay = xi_decay
        self.constraints_method = constraints_method
        self.reducer = reducer
        self.n_components = n_components
        self.inverter_transform = inverter_transform
        self.verbose = verbose

    def optimize(self):

        # *************** Main program ********************
        # Ignore warnings
        warnings.filterwarnings("ignore")
        #
        dims, n_x, n_d, n_c, n_y, x_l, x_u, y_v, dis_val, cat_val, enc_cat, names = Space(self.domain)
        Errors(n_x, n_y, self.design, self.surrogate, self.constraints_method, self.acquisition_function)
        problem_type = Problem_type(n_x, n_y)
        # Get constraints
        const, n_const = Get_constraints(self.constraints, self.constraints_method)
        # If an initial database in not provided, create using the design of experiments
        x, z = Get_x_and_z(self.function, self.x_0, self.z_0, x_l, x_u, y_v, n_x, n_y, n_c, dims, enc_cat, self.p_design, self.design, problem_type)
        # Find reducer and perform dimensionality reduction
        if self.reducer is None:
            if (problem_type == "Continuous") and (n_x < 5):
                self.reducer = "No"
                reducer_trained = None
                dims_red = dims
                x_red = x
            else:
                self.reducer, reducer_trained, dims_red = Find_reducer(x, n_x, dims, self.n_components, problem_type)
                x_red = Reduce(x, n_x, dims, problem_type, reducer_trained)
        elif self.reducer == "No":
            reducer_trained = None
            dims_red = dims
            x_red = x
        else:
            dims_red = self.n_components
            x_red, reducer_trained = Train_reducer(x, n_x, dims, problem_type, dims_red, self.reducer)
        # Reduce bounds
        x_l_red, x_u_red = Red_bounds(x_l, x_u, y_v, n_x, n_y, dims, dims_red, problem_type, reducer_trained)
        # Find inverter
        if self.inverter_transform == "Yes":
            if self.reducer.__module__ == 'prince.mca' or self.reducer.__module__ == 'prince.famd':
                raise ValueError("Reducer module has no inverse_transform")
            else:
                inverter = reducer_trained
        elif self.inverter_transform == "No":
            if self.reducer == "No":
                inverter = None
            else:
                inverter = Find_inverter(x, x_red, n_x, n_y, dims, problem_type)
                inverter = Train_inverter(x, x_red, dims, inverter)
        else:
            pass
        # Generate mesh in the reduced space
        p_mesh_red = Points_mesh(dims_red)
        mesh_red = x_Generator(x_l_red, x_u_red, 0, 0, 0, dims_red, p_mesh_red, "Continuous", "Mesh")
        p_mesh_0 = p_mesh_red
        p_mesh_red, mesh_red = [p_mesh_red], [mesh_red]
        # Tuple of the bounds
        bounds = Bounds(x_l_red, x_u_red, dims_red)
        # Spawning processes
        jobs = Num_jobs(self.n_jobs)
        # Select covariance
        kernel_ = Get_kernel(x_red, z, dims_red, self.kern_discovery, self.kern_discovery_evals, self.surrogate, self.kernel)
        # Eval and train constraints
        if self.constraints is None:
            g_hat, models_const = None, None
        else:
            x_eval = Data_eval(x, n_c, dims, enc_cat)
            g_hat = Eval_const(x_eval, const, n_const, self.constraints_method)
            models_const = Train_models_const(x_red, g_hat, n_const, self.constraints_method)
        # Initialize parameters of acquisition function
        af_params = AF_params(z, self.xi, self.xi_decay, self.max_iter, self.acquisition_function, self.sense)
        #
        x_symb = sp.Matrix(sp.symbols('x:' + str(dims_red)))
        x_symb_names = sp.Matrix(sp.symbols('x:' + str(dims)))
        q, q_inc, delta_q = Percentile_q(self.max_iter)
        flag = 0
        rt = []
        #
        x_best, z_best = Best_values(x, z, self.sense)
        af_params["f_best"] = z_best
        # Print results if vervose is active
        if self.verbose == 1:
            header = Print_header(names, x_symb_names, dims)
            print(header)
            x_print, z_print = Print_results(x_best, z_best, n_c, dims, enc_cat)
            print(0, ' ', z_print, *x_print)
        # Cycle
        for ite in range(1, self.max_iter+1):
            # Train surrogate model
            model = Select_model(x_red, z, kernel_, self.surrogate)
            model = Train_model(model, self.n_restarts)
            # Train surrogate model for constraints
            if self.constraints != None:
                x_eval = Data_eval(x, n_c, dims, enc_cat)
                g_hat = Eval_const(x_eval, const, n_const, self.constraints_method)
                models_const = Train_models_const(x_red, g_hat, n_const, self.constraints_method)
            # Find conected elements
            connected_elements, n_elements = Slf(mesh_red, p_mesh_red, dims_red, q, jobs, af_params, self.constraints_method, self.sense, model, models_const, AF)
            # Querry point
            x_red_new = Querry(x_symb, connected_elements, n_elements, bounds, dims_red, af_params, self.constraints_method, self.sense, model, models_const, AF)
            #
            if self.reducer == "No":
                x_new = x_red_new
            else:
                x_new = Inverse(x_red_new, self.inverter_transform, inverter)
            #
            x_eval = Data_eval(x_new, n_c, dims, enc_cat)
            z_new = Eval_fun(x_eval, n_elements, jobs, self.function)
            # 
            if jobs == 1:
                z_new_best = z_new
            else:
                if self.sense == "maximize":
                    z_new_best = np.max(z_new)
                elif self.sense == "minimize":
                    z_new_best = np.min(z_new)
                else:
                    pass
            # Update train data
            x, z = np.vstack((x, x_new)), np.vstack((z, z_new))
            if self.reducer == "No":
                x_red = np.vstack((x_red, x_red_new))
            else:
                x_red, reducer_trained = Train_reducer(x, n_x, dims, problem_type, dims_red, self.reducer)
                if self.inverter_transform == "No":
                    inverter = Train_inverter(x, x_red, dims, inverter)
                elif self.inverter_transform == "Yes":
                    inverter = reducer_trained
                else:
                    pass
            # 
            if self.sense == "maximize":
                if z_new_best > z_best:
                    flag = 1
            elif self.sense == "minimize":
                if z_new_best < z_best:
                    flag = 1
            # Update parameters
            if flag == 1:
                if self.reducer != "No":
                    connected_elements_inv = [Inverse(connected_elements[i], self.inverter_transform, inverter) for i in range(n_elements)]
                    connected_elements = [Reduce(connected_elements_inv[i], n_x, dims, problem_type, reducer_trained) for i in range(n_elements)]
                mesh_red, p_mesh_red = Up_mesh(connected_elements, n_elements, p_mesh_0, dims_red)
                af_params["xi"] *= af_params["xi_decay"]
                q_inc += delta_q
                q = int(round(q_inc/5.0)*5.0)
            # Update
            rt.append(Regret(z_new, np.array(x_red_new), n_elements, model))
            x_best, z_best = Best_values(x, z, self.sense)
            af_params["f_best"] = z_best
            flag = 0
            # Print results
            if self.verbose == 1:
                x_print, z_print = Print_results(x_best, z_best, n_c, dims, enc_cat)
                print(ite, ' ', z_print, *x_print)

        res = Create_results(x_best, z_best, x, z, x_l, x_u, dims, self.max_iter, self.p_design, self.design, af_params, self.constraints_method, rt, models_const, model)

        return res


# *******************************************************
# Import libraries

import time
import numpy as np
import multiprocess as mp
from scipy.stats import qmc
from sklearn.preprocessing import OrdinalEncoder
from GPy.kern import RBF
from ..utils.Aux import Data_eval
from ..utils.Models import Kernel_discovery

# *******************************************************

def Space(domain):

    """ 
    Gets an array for the lower and upper bounds of the continuous variables, and a tuple for the discrete variables.
    """

    dims = len(domain)
    n_x = 0
    n_d = 0
    n_c = 0
    x_l = []
    x_u = []
    dis_val = []
    cat_val = []
    enc_cat = []
    names = []

    for i in range(dims):
        try:
            names.append(domain[i]['name'])
        except:
            names = 0
        if domain[i]['type'] == "continuous":
            n_x += 1
            x_l.append(domain[i]['domain'][0])
            x_u.append(domain[i]['domain'][1])
        elif domain[i]['type'] == "integer":
            n_d += 1
            dis_val.append(domain[i]['domain'])
        elif domain[i]['type'] == "categorical":
            n_c += 1
            enc_cat.append(OrdinalEncoder())
            X = domain[i]['domain']
            X = [[(X[i])] for i in range(len(X))]
            X_trans = tuple(enc_cat[n_c-1].fit_transform(X).reshape(-1).astype(int))
            cat_val.append(X_trans)
        else:
            pass
            
    x_l, x_u = np.array(x_l), np.array(x_u)
    n_y = n_d + n_c
    y_v = [item for sublist in [dis_val, cat_val] for item in sublist]
    if n_d == 0:
        dis_val = 0
    if n_c == 0:
        enc_cat = None
        cat_val = 0
    if names == 0:
        names = None

    return dims, n_x, n_d, n_c, n_y, x_l, x_u, y_v, dis_val, cat_val, enc_cat, names

# *******************************************************

def Problem_type(n_x, n_y):

    """ 
    Returns the type of problem depending of the characteristics of the inputs:
    * Continuous if there are only continuous variables
    * Discrete if there are only discrete variables
    * Mixed if there are continuous and discrete variables
    """
    
    if n_x > 0 and n_y == 0:
        problem_type = "Continuous"
    elif n_y > 0 and n_x == 0:
        problem_type = "Discrete"
    elif n_x > 0 and n_y > 0:
        problem_type = "Mixed"
    else:
        pass

    return problem_type

# *******************************************************

def Get_constraints(constraints, constraints_method):

    "Transforms tuple of constraints into list"

    if constraints is None:
        const, n_const = None, None
    else:
        n_const = len(constraints)
        if constraints_method == "PoF":
            symbols = ['<=', '>=']
            const = []
            for i in range(n_const):
                for s in symbols:
                    try:
                        index_const = constraints[i]['constraint'].index(s)
                        const.append(constraints[i]['constraint'][:index_const-1])
                    except:
                        pass
        elif constraints_method == "GPC":
            const = constraints
    
    return const, n_const

# *******************************************************

def x_Generator(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, design_type):

    def Random_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type):

        def Bounds_y(y_v, n_y):

            y_l = []
            y_u = []
            flag = int(all([(np.diff(y_v[i]) == 1).all() for i in range(n_y)]))
            
            for i in range(n_y):
                if flag == 1:
                    y_l.append(y_v[i][0])
                    y_u.append(y_v[i][-1]+1)
                elif flag == 0:
                    y_l.append(0)
                    y_u.append(1)
                else:
                    pass
            return y_l, y_u, flag

        def X_rand(x_l, x_u, dims, points):

            return np.random.uniform(x_l, x_u, size=(points, dims))
        
        def Y_rand(y_v, n_y, points):

            y_l, y_u, flag = Bounds_y(y_v, n_y)

            if flag == 1:
                y_rand = np.random.randint(y_l, y_u, size=(points, n_y))
            elif flag == 0:
                size_y = [len(y_v[i]) for i in range(n_y)]
                l = [np.repeat(1/size_y[i], size_y[i]) for i in range(n_y)]
                sample = [np.random.choice(y_v[i], p=l[i], size=(points, 1)).reshape(-1) for i in range(n_y)]
                y_rand = np.array(sample).T
            else:
                pass
            
            return y_rand

        if problem_type == "Continuous":
            variables = X_rand(x_l, x_u, dims, points)
        elif problem_type == "Mixed":
            x_variables = X_rand(x_l, x_u, n_x, points)
            y_variables = Y_rand(y_v, n_y, points)
            variables = np.hstack((x_variables, y_variables))
        elif problem_type == "Discrete":
            variables = Y_rand(y_v, dims, points)
        else:
            pass

        return variables
    
    def QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, design_type):

        def Bounds_y(y_v, n_y):

            y_l = []
            y_u = []
            flag = int(all([(np.diff(y_v[i]) == 1).all() for i in range(n_y)]))
            
            for i in range(n_y):
                if flag == 1:
                    y_l.append(y_v[i][0])
                    y_u.append(y_v[i][-1])
                elif flag == 0:
                    y_l.append(0)
                    y_u.append(1)
                else:
                    pass
            return y_l, y_u, flag

        def X_qmc(x_l, x_u, points, method):
                    
            sample = method.random(n=points)

            return qmc.scale(sample, x_l, x_u)

        def Y_qmc(y_v, n_y, points, method):

            def Assign_y(y_v, n_y, y_norm, points):

                size_y = [len(y_v[i]) for i in range(n_y)]
                l = [[i/size_y[j] for i in range(size_y[j]+1)] for j in range(n_y)]
                for i in range(len(l)):
                    l[i][-1] = l[i][-1] + 0.0001
                y_new = np.empty([points, n_y])

                for i in range(n_y):
                    for j in range(len(y_norm[:,i])):
                        for k in range(len(l[i])-1):
                            if l[i][k] <= y_norm[j,i] < l[i][k+1]:
                                y_new[j,i] = y_v[i][k]
                                break
                
                return y_new
            
            y_l, y_u, flag = Bounds_y(y_v, n_y)
            if flag == 1:
                y_qmc = method.integers(l_bounds=y_l, u_bounds=y_u, n=points, endpoint=True)
            elif flag == 0:
                y_norm = X_qmc(y_l, y_u, points, method)
                y_qmc = Assign_y(y_v, n_y, y_norm, points)        

            return y_qmc

        def Select_method(design_type, dims):
            
            if design_type == "LHS":
                method = qmc.LatinHypercube(d=dims)
            elif design_type == "Sobol":
                method = qmc.Sobol(d=dims)
            elif design_type == "Halton":
                method = qmc.Halton(d=dims)
            else:
                pass

            return method

        if problem_type == "Continuous":
            method = Select_method(design_type, dims)
            variables = X_qmc(x_l, x_u, points, method)
        elif problem_type == "Mixed":
            method = Select_method(design_type, n_x)
            x_variables = X_qmc(x_l, x_u, points, method)
            method = Select_method(design_type, n_y)
            y_variables = Y_qmc(y_v, n_y, points, method)
            variables = np.hstack((x_variables, y_variables))
        elif problem_type == "Discrete":
            method = Select_method(design_type, n_y)
            variables = Y_qmc(y_v, n_y, points, method)
        else:
            pass

        return variables
    
    def Mesh_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type):

        def X_mesh(lower_bound, upper_bound, dims, points):

            lists = [np.linspace(lower_bound[i], upper_bound[i], points) for i in range(dims)]
            mesh = np.meshgrid(*lists)
            
            return np.array(mesh).T.reshape(-1, dims)
        
        def Bounds_y(n_y):

            y_l = np.zeros(n_y)
            y_u = np.ones(n_y)
            
            return y_l, y_u

        def Assign_y(y_v, n_y, y_norm, points):

                size_y = [len(y_v[i]) for i in range(n_y)]
                l = [[i/size_y[j] for i in range(size_y[j]+1)] for j in range(n_y)]
                for i in range(len(l)):
                    l[i][-1] = l[i][-1] + 0.0001
                y_new = np.empty([points, n_y])
                for i in range(n_y):
                    for j in range(len(y_norm[:,i])):
                        for k in range(len(l[i])-1):
                            if l[i][k] <= y_norm[j,i] < l[i][k+1]:
                                y_new[j,i] = y_v[i][k]
                                break
                
                return y_new

        if problem_type == "Continuous":
            variables = X_mesh(x_l, x_u, dims, points)
        elif problem_type == "Mixed":
            y_l, y_u = Bounds_y(n_y)
            all_vars = X_mesh(np.append(x_l, y_l), np.append(x_u, y_u), dims, points)
            x_variables = all_vars[:, :n_x]
            y_norm = all_vars[:, n_x:]
            y_variables = Assign_y(y_v, n_y, y_norm, y_norm.shape[0])
            variables = np.hstack((x_variables, y_variables))
        elif problem_type == "Discrete":
            y_l, y_u = Bounds_y(n_y)
            y_norm = X_mesh(y_l, y_u, n_y, points)
            variables = Assign_y(y_v, n_y, y_norm, y_norm.shape[0])
        else:
            pass

        return variables

    # Main program
    if design_type == "random":
        variables = Random_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type)
    elif design_type == "LHS":
        variables = QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, "LHS")
    elif design_type == "Sobol":
        variables = QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, "Sobol")
    elif design_type == "Halton":
        variables = QMC_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type, "Halton")
    elif design_type == "Mesh":
        variables = Mesh_design(x_l, x_u, y_v, n_x, n_y, dims, points, problem_type)
    else:
        pass
    
    return variables

# ******************************************************* 

def Get_x_and_z(fun, x_0, z_0, x_l, x_u, y_v, n_x, n_y, n_c, dims, enc_cat, p_design, design, problem_type):

# *************************

    def Times_fun(fun, x):

        """ 
        Generates the number of points of the mesh or grid, depending on the cost of the evaluation of the function, and the dimensions of the problem
        """
        
        start = time.time()
        f_eval = fun(x.reshape(1,-1))
        end = time.time()
        
        return (end - start), f_eval

# *************************

    def Points_initial_design(times, dims, design, c_param = 50):

        """ 
        Generates the number of initial points of the design
        """

        if times <= 1:
            exp_param = 0.25
        else: 
            exp_param = 0.95
        points_D = int(c_param - c_param/(1+(1/times)**exp_param))
        # Adjust points if design is Mesh or Sobol. 
        if design == "Mesh":
            points_D = int(np.ceil(points_D)**(1/dims))
        elif design == "Sobol":
            points_D = int(np.ceil(np.sqrt(points_D))**2)
        else:
            pass
        if points_D < 3:
            points_D = int(3)

        return points_D
    
    if x_0 is None:
        # Evaluate an arbitrary point to determine the computation time of the function
        x_trial = x_Generator(x_l, x_u, y_v, n_x, n_y, dims, 1, problem_type, "random")
        x_eval = Data_eval(x_trial, n_c, dims, enc_cat)
        times, z_trial = Times_fun(fun, x_eval)
        if p_design == None:
            p_design = Points_initial_design(times, dims, design)
        x_0 = x_Generator(x_l, x_u, y_v, n_x, n_y, dims, p_design, problem_type, design)
        x_eval = Data_eval(x_0, n_c, dims, enc_cat)
        z_0 = fun(x_eval).reshape(-1,1)
        x, z = np.vstack((x_0, x_trial)), np.vstack((z_0, z_trial))
    else:
        x = x_0
        if z_0 is None:
            x_eval = Data_eval(x_0, n_c, dims, enc_cat)
            z = fun(x_eval).reshape(-1,1)
        else:
            p_design = len(z_0)
            z = z_0
    
    return x, z

# *******************************************************

def Bounds(x_l, x_u, dims):

    bnds = np.sort(np.array([[x_l[i], x_u[i]] for i in range(dims)]))
    return tuple(map(tuple, bnds))

# *******************************************************

def Get_kernel(x_red, z, dims_red, kern_discovery, kern_discovery_evals, surrogate, kernel):

    if kern_discovery == "yes":
        model = Kernel_discovery(x_red, z, dims_red, surrogate, kern_discovery_evals)
        kernel_ = model.kern
    elif kern_discovery == "no" and kernel is None:
        kernel_ = RBF(input_dim=dims_red, variance=1.0, lengthscale=1.0)
    else:
        kernel_ = kernel

    return kernel_

# *******************************************************

def Points_mesh(dims, r1=10):

    """ 
    Generates the points mesh
    """
    points = int(np.ceil(2**(r1/dims)))
    if points < 3:
        points = 3

    return points

# *******************************************************

def Percentile_q(max_iter):

    q0 = 25
    qf = 75
    delta_q = (qf-q0)/max_iter
    q_inc = q0
    q = q0

    return q, q_inc, delta_q

# *******************************************************

def Num_jobs(n_jobs):

    if n_jobs == -1:
        jobs = mp.cpu_count()
    elif n_jobs == None:
        jobs = 1
    else:
        jobs = n_jobs
    
    return jobs

# *******************************************************

def AF_params(z, xi, xi_decay, iters, AF_name, sense):

    if xi_decay == "yes":
        xi_decay = (1/xi)**(1/iters)
    elif xi_decay == "no":
        xi_decay = 1
    else:
        pass

    if sense == "maximize":
        f_best = np.max(z)
    elif sense == "minimize":
        f_best = np.min(z)
    else:
        pass

    return {'xi': xi, 'xi_decay': xi_decay, 'f_best': f_best, 'AF_name': AF_name}